{-# LANGUAGE NamedFieldPuns #-}
module Kenken where

import           Data.Char     (ord)
import qualified Data.Map      as M
import qualified Data.Set      as S
import qualified Data.List     as L
import           Data.List.Split (splitOn)
import           Control.Monad (guard, foldM, msum, fmap)
import           System.IO     (IOMode (ReadMode), hGetContents, openFile)

type Address = (Int, Int)
data Constraint = Constraint {members :: [Address], op :: Operation, target :: Int} deriving (Show, Eq)
data Operation = Add | Sub | Mul | Div | Asrt deriving (Show, Eq)

data Partial = Partial {state :: M.Map Address [Int]
                       , pPeers :: M.Map Address (S.Set Address)
                       , pUnits :: M.Map Address Constraint
                       } deriving (Show, Eq)


crossProd :: [a] -> [b] -> [(a, b)]
crossProd as bs = (,) <$> as <*> bs

translate :: String -> Address
translate (r : c : []) = (ord r - 64, read $ c : "")
translate _            = undefined

symToOp :: String -> Operation
symToOp "+"   = Add
symToOp "-"   = Sub
symToOp "*"   = Mul
symToOp "\\/" = Div
symToOp "!"   = Asrt
symToOp _   = undefined

readKenkenFile :: String -> IO (String)
readKenkenFile fname = do
                    h <- openFile fname ReadMode
                    hGetContents h

showUnits :: Address -> IO ([Constraint], Int) -> IO (Maybe Constraint)
showUnits addr gameIO = do game <- gameIO
                           let us = units game
                           return $ M.lookup addr us

showPeers :: Address -> IO ([Constraint], Int) -> IO (Maybe (S.Set Address))
showPeers addr gameIO = do game <- gameIO
                           let ps = peers . units $ game
                           return $ M.lookup addr ps

readKenken :: String -> ([Constraint], Int)
readKenken kkStr = let spec : groups = splitOn ";" kkStr
                       translated = do 
                                      (op : target : group) <- map words groups
                                      return $ Constraint {op = symToOp op,
                                                  target = read target,
                                                  members = map translate group}
                       [_, size] = words spec in 
                    (translated, read size)

units :: ([Constraint], Int) -> M.Map Address Constraint
units (constraints, _) = M.fromList associations
          where associations = [(a, c) | 
                                c <- constraints,
                                a <- members c
                               ]

peers :: M.Map Address Constraint -> M.Map Address (S.Set Address)
peers m = M.fromList [(addr, neighbors) | (addr, c) <- M.toList m,
                                     let duplicateSet = S.fromList $ members c
                                         neighbors = duplicateSet `S.difference` S.singleton addr]

parsePuzzle :: ([Constraint], Int) -> Partial
parsePuzzle cs@(constraints, size) = Partial {state, pPeers, pUnits} where
                                  pUnits = units cs
                                  pPeers = peers pUnits
                                  state = M.fromList [(a, [1..size]) | c <- constraints, a <- members c]

eliminate :: Address -> Int -> Partial -> Maybe Partial
eliminate a v p = 
  do removed <- remove a v p
     unitPropagated <- propagateUnit removed a
     return unitPropagated


assign :: Address -> Int -> Partial -> Maybe Partial
assign a v p =
  do assigned <- foldM (\partial value -> eliminate a value partial) p toRemove
     propagateSet assigned a
  where toRemove = filter (/= v) $ (state p) M.! a

remove :: Address -> Int -> Partial -> Maybe Partial
remove adr val p = 
  do candidates <- M.lookup adr $ state p
     let reducedCands = L.delete val candidates
         len = L.length reducedCands
     guard (len /= 0)
     return p{state=M.insert adr reducedCands $ state p}

propagateSet :: Partial -> Address -> Maybe Partial
propagateSet p@Partial{state} adr@(row, col) = 
  let  [x] = state M.! adr
       size = fst $ fst $ M.findMax state
       setPeers = [(row, c) | c <- [1..size], c /= col] ++ 
                  [(r, col) | r <- [1..size], r /= row]
       in foldM (\partial peer -> eliminate peer x partial) p setPeers

propagateUnit :: Partial -> Address -> Maybe Partial
propagateUnit p@Partial{pUnits} adr =
  let constraint = pUnits M.! adr in 
  case op $ constraint of
    Add -> propagateAdd p constraint
    Mul -> propagateMul p constraint
    Sub -> propagateSub p constraint
    Div -> propagateDiv p constraint
    _ -> Just p

existsSum ::  Int -> [[Int]] -> Bool
existsSum target []         = target == 0
existsSum target (hd:rest)  = or [existsSum (target - choice) $ rest | choice <- L.reverse hd, choice <= target]

existsProd ::  Int -> [[Int]] -> Bool
existsProd target []         = target == 1
existsProd target (hd:rest)  = or [existsProd (target `div` choice) $ rest | choice <- L.reverse hd, target `mod` choice == 0] 

propagateAdd :: Partial -> Constraint -> Maybe Partial
propagateAdd p@Partial{state} constraint =
  let prs = members constraint
      t = target constraint
      possibilities = [(peer, state M.! peer) | peer <- prs]
      fixed = L.filter (\(_,l) -> L.length l == 1) $ possibilities
      free = L.filter (\(_,l) -> L.length l > 1) $ possibilities
      freeVals = map snd free
      fixedVals = map snd fixed
      newTarget = t - (sum $ msum $ fixedVals) in
  if not $ existsSum newTarget $ L.reverse $ L.sort $ freeVals
  then Nothing
  else return p

propagateMul :: Partial -> Constraint -> Maybe Partial
propagateMul p@Partial{state} constraint =
  let prs = members constraint
      t = target constraint
      possibilities = [(peer, state M.! peer) | peer <- prs]
      fixed = L.filter (\(_,l) -> L.length l == 1) $ possibilities
      free = L.filter (\(_,l) -> L.length l > 1) $ possibilities
      fixedVals = map snd fixed
      freeVals = map snd free
      newTarget = t `div` (product $ msum $ fixedVals) in 
  if not $ existsProd newTarget $ freeVals
  then Nothing
  else return p



propagateSub:: Partial -> Constraint -> Maybe Partial
propagateSub p@Partial{state} constraint =
  let prs = members constraint
      t =  target constraint
      possibilities = [(peer, state M.! peer) | peer <- prs]
      fixed = L.filter (\(_,l) -> L.length l == 1) $ possibilities
      free = L.filter (\(_,l) -> L.length l > 1) $ possibilities
      largest:rest = L.reverse $ L.sort $ msum $ map snd fixed
      result = largest - (sum rest) in
  case free of
     [] -> if result == t then return p else Nothing
     _  -> return p

propagateDiv:: Partial -> Constraint -> Maybe Partial
propagateDiv p@Partial{state} constraint =
  let prs = members constraint
      t = target constraint
      possibilities = [(peer, state M.! peer) | peer <- prs]
      fixed = L.filter (\(_,l) -> L.length l == 1) $ possibilities
      free = L.filter (\(_,l) -> L.length l > 1) $ possibilities
      largest:rest = L.reverse $ L.sort $ msum $ map snd fixed
      result = largest `div` (sum rest) in
  case free of
     [] -> if result == t then return p else Nothing
     _  -> return p





reduce :: Partial -> Address -> Int -> Maybe Partial
reduce p@Partial{state} (row, col) value =
  let size = fst $ fst $ M.findMax state
      rowUnit = [(row, c) | c <- [1..size]]
      colUnit = [(r, col) | r <- [1..size]] in 
  do let rowCandidates = [a | a <- rowUnit, value `elem` state M.! a]
     rowReduced <- case rowCandidates of
                        [] -> Nothing
                        [adr] -> assign adr value p
                        _ -> return p
     let colCandidates = [a | a <- colUnit, value `elem` state M.! a]
     case colCandidates of
          [] -> Nothing
          [adr] -> assign adr value rowReduced
          _ -> return rowReduced


applyAssertions  :: Partial -> Maybe Partial
applyAssertions p@Partial{pUnits} = 
  let asserts = M.filter (\c -> op c == Asrt) pUnits
      assocList = L.map (\(adr, c) -> (adr, target c)) $ M.toList asserts in
  foldM (\partial (a, val) -> assign a val partial) p assocList

solve :: String -> Maybe Partial
solve s = do p <- (applyAssertions . parsePuzzle . readKenken) s
             search p

search :: Partial -> Maybe Partial
search p@Partial{state} = 
  let remaining = M.filter (\l -> (tail l) /= []) state in 
  if remaining == M.empty then return p
  else let  addresses = (M.toList remaining) :: [(Address, [Int])]
            assocComp = \(_, c1) (_,c2) -> (L.length c1) `compare` (L.length c2)
            (nextAdr, candidateVals) = L.minimumBy assocComp addresses
            searchAssign = \v -> assign nextAdr v p >>= \p2 -> search p2
            assigned = map searchAssign candidateVals in
       do result <- msum assigned
          checkSolution result

checkSolution :: Partial -> Maybe Partial
checkSolution p@Partial{state, pUnits} =
  if allCorrect then return p else Nothing
  where allCorrect = and $ map satisfied $ map snd $ M.toList pUnits 
        satisfied Constraint{members, op, target} =
          case op of
            Add -> (sum [head $ state M.! member | member <- members]) == target
            Mul -> (product [head $ state M.! member | member <- members]) == target
            _ -> True

showSolution :: Maybe Partial -> IO ()
showSolution (Just sol) = do putStrLn $ show [v | (_, [v]) <- M.toAscList $ state sol]
showSolution (Nothing) = do putStrLn $ "No Solution Found"

parseSolutionString :: String -> [Int]
parseSolutionString "" = []
parseSolutionString (',':rest) = parseSolutionString rest
parseSolutionString (x:rest) = (read (x:"")) : (parseSolutionString rest)

readSolutionFile :: String -> IO [Int]
readSolutionFile sFile = do sol <- readFile sFile
                            return $ parseSolutionString sol


