import Test.Hspec
import Test.QuickCheck
import Control.Exception (evaluate)
import qualified Kenken as K
import qualified Data.Map      as M
import qualified Data.Set      as S
import           Data.Maybe (fromMaybe, isJust)
import           Control.Monad (msum)



main :: IO ()
main = hspec $ do
  describe "Kenken" $ do
    describe "translate" $ do

      it "A1 is top left" $ do
        (K.translate "A1") `shouldBe` ((1, 1) :: K.Address)

      it "translate puts letter as col" $ do
        (K.translate "E9") `shouldBe` ((5, 9) :: K.Address)

    describe "readKenken" $ do
      it "simple example" $ do
        (K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2") `shouldBe`
          ([K.Constraint {K.members = [(1,1), (2,1)], K.op = K.Add, K.target = 3},
            K.Constraint {K.members = [(1,2), (2,2)], K.op = K.Sub, K.target = 1}],2)

    describe "units" $ do
      it "simple example units of (1,1)" $ do
        let simplePuzzle = K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2"
            constraintMap = K.units simplePuzzle
            actualUnits = M.lookup (1,1) $ constraintMap in 
            actualUnits `shouldBe`
              Just K.Constraint {K.members = [(1,1), (2,1)], K.op = K.Add, K.target = 3}
      it "simple example units of (1,2)" $ do
        let simplePuzzle = K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2"
            constraintMap = K.units simplePuzzle
            actualUnits = M.lookup (1,2) $ constraintMap in 
            actualUnits `shouldBe`
              Just K.Constraint {K.members = [(1,2), (2,2)], K.op = K.Sub, K.target = 1}
      it "simple example units of (2,1)" $ do
        let simplePuzzle = K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2"
            constraintMap = K.units simplePuzzle
            actualUnits = M.lookup (2,1) $ constraintMap in 
            actualUnits `shouldBe`
              Just K.Constraint {K.members = [(1,1), (2,1)], K.op = K.Add, K.target = 3}
      it "simple example units of (2,2)" $ do
        let simplePuzzle = K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2"
            constraintMap = K.units simplePuzzle
            actualUnits = M.lookup (2,2) $ constraintMap in 
            actualUnits `shouldBe`
              Just K.Constraint {K.members = [(1,2), (2,2)], K.op = K.Sub, K.target = 1}

      it "address not found" $ do
        let simplePuzzle = K.readKenken "#\t2;+\t3\tA1 B1;-\t1\tA2 B2"
            constraintMap = K.units simplePuzzle
            actualUnits = M.lookup (3, 4) $ constraintMap in 
            actualUnits `shouldBe` Nothing

    describe "trying solver" $ do
      it "some 3s" $ do
        let names = ["puzzles/3_" ++ show num ++ ".txt" | num <- [0..10]]
            puzzles = map readFile names
            solved = sequence $ map (\puzzIO -> fmap K.solve $ puzzIO) puzzles
            res = fmap (all isJust) solved in 
            res `shouldReturn` True
      it "some 4s" $ do
        let names = ["puzzles/4_" ++ show num ++ ".txt" | num <- [0..10]]
            puzzles = map readFile names
            solved = sequence $ map (\puzzIO -> fmap K.solve $ puzzIO) puzzles
            res = fmap (all isJust) solved in 
            res `shouldReturn` True
      it "some 5s" $ do
        let names = ["puzzles/5_" ++ show num ++ ".txt" | num <- [0..10]]
            puzzles = map readFile names
            solved = sequence $ map (\puzzIO -> fmap K.solve $ puzzIO) puzzles
            res = fmap (all isJust) solved in 
            res `shouldReturn` True
      it "some 6s" $ do
        let names = ["puzzles/6_" ++ show num ++ ".txt" | num <- [0..10]]
            puzzles = map readFile names
            solved = sequence $ map (\puzzIO -> fmap K.solve $ puzzIO) puzzles
            res = fmap (all isJust) solved in 
            res `shouldReturn` True
      it "some 7s" $ do
        let names = ["puzzles/7_" ++ show num ++ ".txt" | num <- [0..10]]
            puzzles = map readFile names
            solved = sequence $ map (\puzzIO -> fmap K.solve $ puzzIO) puzzles
            res = fmap (all isJust) solved in 
            res `shouldReturn` True

    describe "correctness" $ do
      it "a 3" $ do
        let puzzle = "puzzles/3_58.txt"
            solution = "solutions/3_58.txt"
            actualM = fmap K.solve $ readFile $ puzzle
            expectedM = K.readSolutionFile solution in
            do expected <- expectedM
               actual <- actualM
               let res = msum $ map snd $ M.toList $ K.state $ fromMaybe undefined actual
               res `shouldBe` expected

      it "a 4" $ do
        let puzzle = "puzzles/4_58.txt"
            solution = "solutions/4_58.txt"
            actualM = fmap K.solve $ readFile $ puzzle
            expectedM = K.readSolutionFile solution in
            do expected <- expectedM
               actual <- actualM
               let res = msum $ map snd $ M.toList $ K.state $ fromMaybe undefined actual
               res `shouldBe` expected
      
      it "a 5" $ do
        let puzzle = "puzzles/5_58.txt"
            solution = "solutions/5_58.txt"
            actualM = fmap K.solve $ readFile $ puzzle
            expectedM = K.readSolutionFile solution in
            do expected <- expectedM
               actual <- actualM
               let res = msum $ map snd $ M.toList $ K.state $ fromMaybe undefined actual
               res `shouldBe` expected

      it "a 6" $ do
        let puzzle = "puzzles/6_58.txt"
            solution = "solutions/6_58.txt"
            actualM = fmap K.solve $ readFile $ puzzle
            expectedM = K.readSolutionFile solution in
            do expected <- expectedM
               actual <- actualM
               let res = msum $ map snd $ M.toList $ K.state $ fromMaybe undefined actual
               res `shouldBe` expected

      it "a 7" $ do
        let puzzle = "puzzles/7_58.txt"
            solution = "solutions/7_58.txt"
            actualM = fmap K.solve $ readFile $ puzzle
            expectedM = K.readSolutionFile solution in
            do expected <- expectedM
               actual <- actualM
               let res = msum $ map snd $ M.toList $ K.state $ fromMaybe undefined actual
               res `shouldBe` expected
      
